/******************************************************************************/
/*   IBM Confidential							 */
/*									    */
/*   OCO Source Materials.						    */
/*									    */
/*   (C) Copyright IBM Corp. 2001.					    */
/*									    */
/*   The source code for this program is not published or otherwise	   */
/*   divested of its trade secrets, irrespective of what has been	     */
/*   deposited with the U.S. Copyright Office.				*/
/*									    */
/*   Product ID:							      */
/******************************************************************************/

/**
 * Descriptive Name:  VNurse.java
 *
 * Description:
 *
 *
 *    Modifications:
 *
 *    Date      Author		 Category	     Defect/Feature
 *	      Description
 *
 *  08/08/2001  Dave Been	      Creation
 *
 */

package com.ibm.mobileservices.demo;

import com.ibm.mobileservices.lang.SampleCLPmsg;

/**
 * DB2 Everyplace Virtual Nurse Demo
 *
 *	      This class is the MIDlet run inside the device
 */

import javax.microedition.midlet.*;
import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.ChoiceGroup;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.Gauge;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.Item;
import javax.microedition.lcdui.ItemStateListener;
import javax.microedition.lcdui.TextField;
import javax.microedition.lcdui.Ticker;

import java.io.InputStream;
import java.io.DataInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;

import java.lang.Exception;
import java.io.IOException;
import java.io.EOFException;
import java.util.Hashtable;

import com.ibm.mobileservices.isync.ISync;
import com.ibm.mobileservices.isync.ISyncProvider;
import com.ibm.mobileservices.isync.ISyncConfigStore;
import com.ibm.mobileservices.isync.ISyncService;
import com.ibm.mobileservices.isync.ISyncDriver;
import com.ibm.mobileservices.isync.midp.MIDPISyncProvider;
import com.ibm.mobileservices.isync.midp.TableMetaData;
import com.ibm.mobileservices.isync.midp.ConflictReader;
import com.ibm.mobileservices.isync.sql.Types;
import com.ibm.mobileservices.isync.ISyncException;
import com.ibm.mobileservices.isync.event.ISyncEvent;
import com.ibm.mobileservices.isync.event.ISyncListener;
import com.ibm.mobileservices.isync.ISyncSubscriptionSet;

import com.ibm.mobileservices.isync.debug.Debug;

public class VNurse extends MIDlet
	implements CommandListener, ItemStateListener, ISyncListener, Runnable
{
	private String eventString;		// Debug
	private boolean printedTrace;   // print stack trace once

	private ISyncProvider provider;
	private ISyncService service;
	private ISyncConfigStore config;
	private ISyncDriver isync;		      // Sync Client API instance
	private int syncState = 0;

	private String packetUpSize, packetDownSize;

	private Thread syncThread;

	private Command initAppButton;
	private Command addButton;
	private Command updateButton;
	private Command backButton;
	private Command contactsButton;
	private Command deleteButton;
	private Command exitButton;
	private Command infoButton;
	private Command recordsButton;
	private Command saveButton;
	private Command scheduleButton;
	private Command syncButton;
	private Command viewButton;

	private Command cancelSync;

	private Command[] recDetailButtons;
	private Command[] showSchedButtons;
	private Command[] showInfoButtons;
	private Command[] showContactButtons;
	private Command[] showContactInfoButtons;
	private Command[] showRecordButtons;

	private Form schedForm;
	private Form infoForm;
	private Form recForm;
	private Form recDetailForm;	     // medVerbose set to true



	private Form contactForm;

	private Form contactInfoForm;
	
	private Form addForm;

	private Display display;		// The display for this MIDlet
	private Ticker ticker;		  // Left Scrolling text 
	private Alert errorAlert;
	Form prevForm;
	Form currentForm;
	private Form statusForm;


	private Gauge syncGauge;
	private Form syncForm;
	private TextField syncStateField;

	private String userName;
	private String password;

	private NursesAid aid;

	public VNurse()
	{
		/* 
		 * Set up graphics presentation
		 */

		display	 = Display.getDisplay(this);
		initAppButton   = new Command(SampleCLPmsg.ResetSubs, Command.SCREEN,   2);
		addButton       = new Command(SampleCLPmsg.Add, Command.SCREEN,   2);
		updateButton= new Command(SampleCLPmsg.Update, Command.SCREEN,   2);
		backButton      = new Command(SampleCLPmsg.Back, Command.BACK, 2);
		contactsButton  = new Command(SampleCLPmsg.Contacts, Command.SCREEN, 2);
		deleteButton    = new Command(SampleCLPmsg.Delete, Command.SCREEN, 2);
		exitButton = new Command(SampleCLPmsg.Exit, Command.SCREEN, 2);
		infoButton = new Command(SampleCLPmsg.Info, Command.SCREEN, 1);
		recordsButton   = new Command(SampleCLPmsg.Records, Command.SCREEN, 1);
		saveButton      = new Command(SampleCLPmsg.Save, Command.SCREEN, 2);
		syncButton      = new Command(SampleCLPmsg.Sync, Command.SCREEN, 1);
		scheduleButton  = new Command(SampleCLPmsg.Schedule, Command.SCREEN, 2);
		viewButton      = new Command(SampleCLPmsg.View, Command.SCREEN, 1);
		ticker	  = new Ticker("");

		cancelSync	      = new Command(SampleCLPmsg.Cancel, Command.SCREEN, 1);

		Command[] a = {addButton, updateButton, deleteButton, backButton, exitButton};
		recDetailButtons	= a;
		Command[] b = {infoButton, backButton, exitButton};
		showSchedButtons	= b;
		Command[] c = {recordsButton, contactsButton, backButton};
		showInfoButtons	 = c;
	Command[] d = {viewButton, addButton, updateButton, deleteButton, backButton, exitButton};
		showRecordButtons       = d;

		Command[] e = {infoButton, backButton};
		showContactButtons      = e;
		Command[] f = {backButton};
		showContactInfoButtons      = f;

		syncForm = new Form(SampleCLPmsg.Synchronizing);
		syncGauge = new Gauge(null, true, 100, 0);

		syncForm.append(syncGauge);
		syncStateField = new TextField(null, SampleCLPmsg.Start, 30, TextField.ANY);
		syncForm.append(syncStateField);
		syncForm.addCommand(cancelSync);
		syncForm.setTicker(ticker);
		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println("	VNurse constructor done ");
	}

	/**
	 * Start up the MIDlet by creating the TextBox and associating
	 * the exit command and listener.
	 */
	public void startApp()
	{
		userName = "nurse1";
		password = "nurse1";

		if (Debug.DEBUG && Debug.LEVEL > 5)
			System.out.println("	startApp ");
		String x = getAppProperty("Db2eSyncUserName");
		// System.out.println("user name from jad file " + x);
		if (x != null)
		{
			userName = x;
		}
		x = getAppProperty("Db2eSyncPassword");
		// System.out.println("password name from jad file " + x);
		if (x != null)
			password = x;
		packetUpSize = getAppProperty("PacketUpSize");
		packetDownSize = getAppProperty("PacketDownSize");
		if (packetUpSize == null)
			packetUpSize = "30000";
		if (packetDownSize == null)
			packetDownSize = "30000";

		setupForSync();

		int version = -1;
		try
		{
			version = provider.getVersion();
		}
		catch (ISyncException ise)
		{
			// can't happen in MIDP
		}

		int maj = (version & 0xff000000) >> 24;
		int min = (version & 0x00ff0000) >> 16;
		int mod = (version & 0x0000ff00) >> 8;

		statusForm = new Form(SampleCLPmsg.IBMSync + maj + "." + min + "." + mod);
		statusForm.addCommand(syncButton);
		statusForm.addCommand(initAppButton);
		statusForm.addCommand(scheduleButton);
		statusForm.addCommand(exitButton);
		// statusForm.addCommand(settingsButton);
		// statusForm.addCommand(setSubsButton);
		// statusForm.addCommand(aboutButton);
		statusForm.addCommand(exitButton);

		statusForm.setCommandListener(this);
		statusForm.setTicker(ticker);
		prevForm = statusForm;
		display.setCurrent(statusForm);
		currentForm = statusForm;

		aid = new NursesAid(isync, display, statusForm);

		// go back to start screen
		callShowSchedule(showSchedButtons);

		try
		{
			Image cs = Image.createImage("/ibm.png");
			statusForm.append(cs);
		}
		catch (IOException ioe) {}

		String bdate = null;
		try
		{
			bdate = provider.getBuildDate();
		}
		catch (ISyncException neverHappens)
		{
		}
		if (bdate != null)
		{
			statusForm.append("Build date:");
			statusForm.append("\n" + bdate);
		}

		if (Debug.DEBUG)
			System.out.println(" startApp done");
	}
	
	private ByteArrayOutputStream trcBout = null;
	private PrintStream trcStream = null;
	private void setupForSync()
	{
		if (Debug.DEBUG)
			System.out.println("setupForSync");

		try
		{
			String defaultHost = "127.0.0.1";
			String defaultPort = "9080";

			/*
				Read settings from JAD file
			*/

			String URI = getAppProperty("Db2eSyncURI");

			String x = getAppProperty("Db2eSyncHost");
			if (x != null)
				defaultHost = x;
			x = getAppProperty("Db2eSyncPort");
			if (x != null)
				defaultPort = x;
			if (URI == null)
			{
				URI =  defaultHost;
				if (defaultPort != null)
					URI = URI += ":" + defaultPort;
			}
			provider = MIDPISyncProvider.getInstance();
			Hashtable ht = new Hashtable();
			ht.put("isync.user", userName);
			ht.put("isync.password", password);

			String lang = getAppProperty("Db2eSyncLang");
			if (lang != null && lang.equals("jp"))
				ht.put("isync.encoding", "UnicodeBig");
			ht.put("isync.trace", ISync.TRACE_ON);

			service = provider.createSyncService(URI, ht);
			config = service.getConfigStore(null);
			isync = config.getSyncDriver();

			trcBout = new ByteArrayOutputStream();
			trcStream = new PrintStream(trcBout);
			((MIDPISyncProvider) isync).setTraceStream(trcStream);
		}
		catch (ISyncException neverthrownInMidp)
		{
			System.out.println(" something failed "
				+ neverthrownInMidp.toString());
		}
		try
		{
			/*
				Set the preference which the IBM Sync engine needs
				to perform the synchronization.The preference is
				persistent.
			*/

			isync.setPref(ISync.PREF_PACKET_UP_MAX, packetUpSize);
			isync.setPref(ISync.PREF_PACKET_DOWN_MAX, packetDownSize);

			/*
				Register the callback to the IBM Sync engine for
				event notification if needed by the application.
				The syncListener() is defined below.
			*/

			isync.setSyncListener(this);
		}
		catch (ISyncException ise)
		{
		System.out.println("setSyncListener failed " + ise.toString());
		ise.printStackTrace();
			// ignore because we know what we are doing
		}
	}
	

    public int eventIssued(ISyncEvent syncEvent)
    {
		int eventCode = syncEvent.getEventCode();
		String eventMsg = syncEvent.getSubscriptionSetName();
		String subName = syncEvent.getSubscriptionName();
		int eventType = syncEvent.getEventType();

		Object listenerInfo = syncEvent.getEventInfo();

		Exception e = null;
		ConflictReader cr = null;

// System.out.println("=== eventIssued ss/s '" + eventMsg + "'/'" + subName + "'");

		if (listenerInfo instanceof Exception)
			e = (Exception) listenerInfo;
		else if (listenerInfo instanceof ConflictReader)
			cr = (ConflictReader) listenerInfo;
		// else TBD
		

		if (e != null)
			eventString += " " + eventCode + " " + e.toString();
		else
			eventString += " " + eventCode;
		if (eventMsg != null)
			eventString += ": " + eventMsg;

		if (Debug.DEBUG)
		{
			String msg = "eventIssued ";
			if (e != null)
				msg += eventCode + " " + e.toString();
			else
				msg += eventCode;

			if (eventMsg != null)
				msg += " Msg " + eventMsg;
			msg += " Type " + eventType;

			if (eventType == ISync.EVTTYPE_INFO)
				System.out.println(msg);
			else
				System.out.print(msg);
		}

		if (e != null && !printedTrace)
		{
			if (Debug.DEBUG && Debug.LEVEL > 10)
			{
				System.out.println("=== stack trace start: " + e.toString());
				e.printStackTrace();
				System.out.println("=== stack trace done");
			}
			printedTrace = true;
		}

		if (eventType == ISync.EVTTYPE_CONFLICT)
		{
			if (cr == null)
			{
				System.out.println("Missing conflict reader");
				displayError(eventMsg, e, eventCode);
				return ISync.RTNCB_DONE;
			}
			System.out.println("got ConflictReader");

			String tableName = cr.getTableName();
			System.out.println("table " + tableName);

			byte[] conflictRow = cr.getRejectedRow();
			int op = cr.getOperation();

			System.out.println(SampleCLPmsg.ConflictTable + tableName
				+ " op " + op + " row: " + conflictRow);

			for (int i = 0; i < conflictRow.length; i++)
			{
				String v = Integer.toHexString(conflictRow[i] & 0xff);
				if (v.length() == 1)
					v = "0" + v;
				System.out.print(v + " ");
				if ((i % 16) == 0)
					System.out.println();
			}
			System.out.println();

			ByteArrayInputStream bin = null;
			DataInputStream din = null;
			try
			{
				bin = new ByteArrayInputStream(conflictRow);
				din = new DataInputStream(bin);

				TableMetaData tmd =
				   ((MIDPISyncProvider)isync).getTableMetaDataByName(tableName);

				if (tmd != null)
				{
					int numCols = tmd.getNumCols();

					System.out.println("table " + tableName + " has #cols "
						+ numCols);

					// Dirty byte
					din.readByte();

					String conflictReason = din.readUTF();
					System.out.println("conflictReason " + conflictReason);

					String conflictType = din.readUTF();
					System.out.println("conflictType " + conflictType);

					/*
						Now read the row data for the table
					*/

					for (int c = 0; c < numCols; c++)
					{
						boolean isNullable = tmd.isNullable(c);
						boolean isPk = tmd.isPrimaryKey(c);

						// only PKs conflicted for DELETEs
						if (op == ISync.OP_DELETE && !isPk)
							continue;
						boolean isNull = false;
						if (isNullable)
							isNull = din.readBoolean();
						if (isNull)
						{
							System.out.println("null, ");
							continue;
						}

						// print out data
						readColumn(din, tmd, c);
					}
					System.out.println();
				}
				else
					System.out.println("Cannot find TableMetaData for " +
						tableName);
			}
			catch (Exception ee)
			{
				System.out.println("Processing conflict row failed "
					+ ee.toString());
				// if (Debug.DEBUG && Debug.LEVEL > 8)
					ee.printStackTrace();
			}
			finally
			{
				if (din != null)
				{
					try
					{
						din.close();
					}
					catch (IOException ioe) { }
				}
			}

			displayError(eventMsg, e, eventCode);
			return ISync.RTNCB_DONE;
		}
		/*
			Retry 3 times
		*/
		if (eventType == ISync.EVTTYPE_RETRY)
		{
			int retries = syncEvent.getNumberOfRetries();
			if (retries < 3)
				return ISync.RTNCB_REPLY_YES;
			return ISync.RTNCB_REPLY_NO;
		}
		if (eventType == ISync.EVTTYPE_ERROR)
		{
			// Error Condition
			displayError(eventMsg, e, eventCode);
			System.out.println("  EVENT :  EVENT_ERROR");
			ticker.setString(SampleCLPmsg.SynchronizeFailed);

			// go back to start screen
			callShowSchedule(showSchedButtons);
			return ISync.RTNCB_DONE;
		}
		switch (eventCode)
		{
		case ISync.EVT_INF_SYNC_PROG:
			syncGauge.setValue(syncEvent.getSyncProgress());
			break;
		case ISync.EVT_INF_SYNC_STARTED:
			syncStateField.setString(SampleCLPmsg.SyncStarted);
			break;
		case ISync.EVT_INF_SUBSSET_STARTED:
			if (eventMsg == null)
				eventMsg = "Config";
			ticker.setString(eventMsg);
			break;
		case ISync.EVT_INF_PREP_MSG:
			syncStateField.setString(SampleCLPmsg.PrepareMsg);
			break;
		case ISync.EVT_INF_SEND_MSG:
			syncStateField.setString(SampleCLPmsg.SendingMsg);
			break;
		case ISync.EVT_INF_WAIT_MSG:
			syncStateField.setString(SampleCLPmsg.WaitOnReply);
			break;
		case ISync.EVT_INF_SYNC_CANCELED:
			if (Debug.DEBUG)
				System.out.println("SYNC_CANCELED");
			showAlert(SampleCLPmsg.Sync, SampleCLPmsg.Canceled, null, eventCode);
			ticker.setString("");

			syncStateField.setString(SampleCLPmsg.SyncCanceled);
			display.setCurrent(statusForm);
			currentForm = statusForm;
			break;

		case ISync.EVT_INF_SYNC_SUCCEEDED:
			if (Debug.DEBUG)
				System.out.println(" Sync Finished");

			ticker.setString(SampleCLPmsg.FinishedSynchronize);

			// go back to start screen
			callShowSchedule(showSchedButtons);

			break;
		
		}
		return ISync.RTNCB_DONE;
    }

	private void readColumn(DataInputStream din, TableMetaData tmd, int c)
		throws IOException, Exception /* getType */
	{
		int type = tmd.getType(c);

		switch (type)
		{
		  case Types.DECIMAL:
		  case Types.CHAR:
		  case Types.VARCHAR:
			String val = din.readUTF();
			System.out.print("'" + val + "'");
			break;
		  	

		  case Types.DATE:
		  case Types.TIME:
		  case Types.INTEGER:
		  {
			int i = din.readInt();
			System.out.print("INT/DATE/TIME " + i);
			break;
		  }

		  case Types.SMALLINT:
		  {
			short s = din.readShort();
			System.out.print("SMALLINT " + s);
			break;
		  }

		  case Types.TINYINT:
		  {
			byte b = din.readByte();
			System.out.print("TINYINT " + b);
			break;
		  }

		  case Types.BIGINT:
		  case Types.TIMESTAMP:
		  {
			long dt = din.readLong();
			System.out.print("BIGINT/TSTAMP " + dt);
			break;
		  }

		  default:
			System.out.println("\n readCol unsupported case "
			  		+ type);
		    break;
		}
		System.out.println(", ");
	}
	
	private void callShowRecords(Command[] showRecordButtons)
	{
		recForm = aid.showRecords(recForm, showRecordButtons);
		recForm.setCommandListener(this);
		recForm.setItemStateListener(this);
		prevForm = currentForm;
		currentForm = recForm;
		display.setCurrent(currentForm);
	}
	private void callShowSchedule(Command[] showSchedButtons)
	{
		try
		{
			schedForm = aid.showSchedule(showSchedButtons);
			if (schedForm != null)
			{
				schedForm.setCommandListener(this);
				schedForm.setItemStateListener(this);
				prevForm = currentForm;
				currentForm = schedForm;
				display.setCurrent(currentForm);
				return;
			}

			// go back to start screen
			display.setCurrent(statusForm);
			currentForm = statusForm;
		}
		catch (Exception ee)
		{
			if (Debug.DEBUG)
				System.out.println("showSchedule failed");
			display.setCurrent(statusForm);
			currentForm = statusForm;
		}
	}

	/**
		Private methods 
	 */
	private void displayError(String eventMsg, Exception e, int eventCode)
	{
		if (Debug.DEBUG)
			System.out.println(" CANCEL/FAIL: " + eventMsg);
		// the request received an exception
		// could be PUB_COMPLETE or PUB_INCOMPLETE or PUB_PRECOPY

		if (eventMsg == null)
		{
			eventMsg = SampleCLPmsg.Error;
			if (e != null)
				eventMsg += e.toString();
		}
		showAlert(eventMsg, SampleCLPmsg.SyncError, e, eventCode);

		if (prevForm == null)
		{
			if (Debug.DEBUG)
				System.out.println(" >>>>>>>>> prevForm is null ");
			display.setCurrent(currentForm);
			currentForm = statusForm;
		}
		else
		{
			display.setCurrent(prevForm);
			currentForm = prevForm;
		}
	}

	private void showAlert(String status, String msg, Exception e,
		int eventCode)
	{
		errorAlert = new Alert(status);
		if (e != null)
			errorAlert.setString(e.getMessage());
		else
			errorAlert.setString(msg);

		// DEFAULT IS 2000 (2 seconds)
		// alert timeouts do not work
		if (eventCode == ISync.EVT_CFT_REJECT)
			errorAlert.setTimeout(Alert.FOREVER);
		else
			errorAlert.setTimeout(3000);

		display.setCurrent(errorAlert);

		try
		{
			Thread.sleep(3000);
		} catch (/* Interrupted */ Exception ee) { }
	}

	/**
	 * Pause is a no-op since there are no background activities or
	 * record stores that need to be closed.
	 */
	public void pauseApp()
	{
		if (Debug.DEBUG && Debug.LEVEL > 1)
			System.out.println("pauseApp");
	}

	/**
	 * Destroy must cleanup everything not handled by the garbage collector.
	 * In this case there is nothing to cleanup.
	 */
	public void destroyApp(boolean unconditional)
	{
		System.out.println("destroyApp");
		if (isync != null)
		{
			try
			{
				isync.close();
			}
			catch (Exception e) { }
			isync = null;
		}
	}
	/*
	 * Respond to commands, including exit
	 * On the exit command, cleanup and notify that the MIDlet has been destroyed.
	 */
	public void commandAction(Command c, Displayable s)
	{
		if (Debug.DEBUG && Debug.LEVEL > 10)
		{
			System.out.println(" >>>>>>>>>> CMD ACTION s " + s);
			if (syncForm != null)
				System.out.println(" syncForm " + syncForm);
			if (schedForm != null)
				System.out.println(" schedForm " + schedForm);
			if (recForm != null)
				System.out.println(" recForm " + recForm);
			if (recDetailForm != null)
				System.out.println(" recDetailForm " + recDetailForm);
			if (contactForm != null)
				System.out.println(" contactForm " + contactForm);
			if (contactInfoForm != null)
				System.out.println(" contactInfoForm " + contactInfoForm);
			if (addForm != null)
				System.out.println(" addForm " + addForm);
			if (statusForm != null)
				System.out.println(" statusForm " + statusForm);
			if (currentForm != null)
				System.out.println(" currentForm " + currentForm);
			if (prevForm != null)
				System.out.println(" prevForm " + prevForm);
		}
		if (c == exitButton)
		{
			destroyApp(false);
			notifyDestroyed();
			return;
		}
		if (s == syncForm)
		{
			if (c == cancelSync)
			{
				try
				{
					if (isync != null)
					{
						syncStateField.setString(SampleCLPmsg.Cancelling);
						isync.cancelSync();
					}
				}
				catch (ISyncException ise)
				{
					// can't happen in MIDP
				}

				// go back to start screen
				callShowSchedule(showSchedButtons);
			}
			else if (Debug.DEBUG)
				System.out.println(" unknown cmd in syncForm");
			return;
		}
		if (c == addButton || c == updateButton)
		{
			addForm = aid.modifyRecord(addForm, saveButton,
					backButton, (c == updateButton), (c == updateButton));
			addForm.setCommandListener(this);
			addForm.setItemStateListener(this);
			prevForm = currentForm;
			currentForm = addForm;
			return;
		}
		else if (c == deleteButton)
		{
			aid.deleteRecord();
			// refresh screen with current records
			callShowRecords(showRecordButtons);
			return;
		}
		// statusForm and contactForm
		if (c == scheduleButton)
		{
			callShowSchedule(showSchedButtons);
			return;
		}
		if (s == recForm)
		{
			if (c == backButton)
			{
				currentForm = infoForm;
				display.setCurrent(currentForm);
			}
			else if (c == viewButton)
			{
				recDetailForm = aid.showRecDetails(recDetailForm,
					recDetailButtons);
				recDetailForm.setCommandListener(this);
				display.setCurrent(recDetailForm);
				currentForm = recDetailForm;
			}
			return;
		}
		if (s == recDetailForm)
		{
			if (c == backButton)
			{
				currentForm = recForm;
				display.setCurrent(currentForm);
			}
			return;
		}
		if (s == schedForm)
		{
			if (c == backButton)
			{
				currentForm = statusForm;
				display.setCurrent(currentForm);
			}
			else if (c == infoButton)
			{
				infoForm = aid.showInfo(infoForm, showInfoButtons);
				infoForm.setCommandListener(this);
				prevForm = currentForm;
				currentForm = infoForm;
			}
			return;
		}
		if (s == infoForm)
		{
			if (c == backButton)
			{
				currentForm = schedForm;
				display.setCurrent(currentForm);
			}
			else if (c == recordsButton)
			{
				callShowRecords(showRecordButtons);
			}
			else if (c == contactsButton)
			{
				contactForm = aid.showContacts(contactForm, showContactButtons);
				contactForm.setCommandListener(this);
				contactForm.setItemStateListener(this);
				prevForm = currentForm;
				currentForm = contactForm;
			}
		    return;
		}
		if (s == contactForm)
		{
			if (c == backButton)
			{
				currentForm = infoForm;
				display.setCurrent(currentForm);
			}
			else if (c == infoButton)
			{
				contactInfoForm = aid.showContactInfo(contactInfoForm,
					showContactInfoButtons);
				contactInfoForm.setCommandListener(this);
				// contactInfoForm.setItemStateListener(this);
			}
			return;
		}
		if (s == contactInfoForm)
		{
			if (c == backButton)
			{
				currentForm = contactForm;
				display.setCurrent(currentForm);
			}
			return;
		}
		if (s == addForm)
		{
			if (c == backButton)
			{
				currentForm = prevForm;
				display.setCurrent(currentForm);
			}
			else if (c == saveButton)
			{
				aid.saveRecord();
				callShowRecords(showRecordButtons);
			}
			return;
		}
		if (s == statusForm)
		{
			if (c == syncButton)
			{
				syncGauge.setValue(0);
				prevForm = currentForm;
				display.setCurrent(syncForm);
				syncForm.setCommandListener(this);
				currentForm = syncForm;
				if (Debug.DEBUG)
					System.out.println(" Start new thread ");
				syncThread = new Thread(this);
				syncThread.start();
				return;
			}
			else if (c == initAppButton)
			{
				// reset to REFRESH mode
				try
				{
					if (config != null)
						config.purge();
				}
				catch (Exception ifyoucan) { } 
				return;
			}
			
			if (Debug.DEBUG && Debug.LEVEL > 5)
				System.out.println(" STATUS FORM unhandled button");
			return;
		}
		if (Debug.DEBUG)
			System.out.println("CommandAction: unknown form " + s);
	}

	public void run()
	{
		printedTrace = false;
		eventString = "events: ";
		for (int i = statusForm.size() - 1; i >= 1; i--)
			statusForm.delete(i);

		try
		{
			if (Debug.DEBUG)
				System.out.println("run: start sync");

			long start = System.currentTimeMillis();

			ticker.setString(SampleCLPmsg.BeginSynchronize);
			
			if (Debug.DEBUG && Debug.LEVEL > 5)
				System.out.println("run: SYNC");

			int rc = syncState = isync.sync();

			switch (rc)
			{
			  case ISync.RTN_SUCCEEDED:
				reportSyncStatus("Synchronization succeeded "
					+ eventString);
				break;

			  case ISync.RTN_CANCELED:
				reportSyncStatus("Synchronization canceled "
				 	+ eventString);      
				break;

			  default: 
				reportSyncStatus("Synchronization failed "
				 	+ eventString);        
				break;
			}        

			if (Debug.PERF_DEBUG && Debug.PERF_LEVEL > 3)
			{
				long end = System.currentTimeMillis();
				if (Debug.PERF_DEBUG)
					System.out.println(" !!!!!!!!!! Total sync time (millis): " 
						+ (end - start));
				start = end;
			}
			if (Debug.DEBUG)
			{
				if (syncState == ISync.RTN_CANCELED)
					System.out.println(" run: SYNC CANCELED");
				else
					System.out.println("run: sync OK");
			}
		}
		catch (ISyncException iex)
		{
			reportSyncStatus("Exception Code: "
				+ iex.getCode() + " "
				+ iex.toString() + ", Event codes: " + eventString);	
		}
		catch (Exception e)
		{
			if (Debug.DEBUG)
				System.out.println(" sync method failed " + e.toString());
			if (Debug.DEBUG && Debug.LEVEL > 8)
				e.printStackTrace();

			reportSyncStatus("Exception : "
				+ e.toString() + ", Event codes: " + eventString);	
		}
		catch (Throwable th)
		{
			/*
				There are conditions such as opening an RMS store that
				might fail if, for example, the phone voice notes fills
				up, causing a RuntimeException to be thrown (see "Nextel
				J2ME Developer's Guide for the i85s and i50sx" document).
			*/
			System.gc();
			displayError(th.toString(), null, 0);
		}
		callShowSchedule(showSchedButtons);
	}
	private void reportSyncStatus(String msg)
	{
		System.out.println("reportSyncStatus " + msg);
		if (statusForm != null)
			statusForm.append(msg);

		// in case we aren't on opening screen
		if (currentForm != statusForm)
			currentForm.append(msg);

		// if (trcStream != null)
		// {
		// 	System.out.println("Have trcStream");
		// 	String trace = new String(trcBout.toByteArray());
		// 	System.out.println(trace);
		// }
	}

	/**
	 * Handle Choice Group, etc state change messages
	 */
	public void itemStateChanged(Item item)
	{
		if (Debug.DEBUG)
			System.out.println("itemStateChanged: " + item);

		if (item instanceof ChoiceGroup)
		{
			ChoiceGroup cg = (ChoiceGroup) item;
			int index = cg.getSelectedIndex();
			aid.checkStateChanged(cg, index);
		}
	}

}
